use std::{u8, i8, f32};
use std::ops::{Add, Mul};
use std::iter::{Iterator, Sum};

pub type Sample = f32;
#[derive(Debug)]
pub struct Frame {
    pub l: Sample,
    pub r: Sample,
}

fn mono_pan_amplitudes(angle: f32) -> (f32, f32) {
    // L_amp = √2/2 (cosθ - sinθ)
    // R_amp = √2/2 (cosθ + sinθ)

    let l = (2.0_f32).sqrt()/2.0 * (angle.cos() - angle.sin());
    let r = (2.0_f32).sqrt()/2.0 * (angle.cos() + angle.sin());
    (l, r)
}

pub fn from_bitpan(angle: i8) -> f32 {
    const MAX_ANGLE: f32 = 0.7853982; // +/- 45 deg

    (angle as f32) / (i8::max_value() as f32) * MAX_ANGLE
}

pub fn from_bitamp(amp: u8) -> f32 {
    (amp as f32) / (u8::max_value() as f32)
}

impl Frame {
    pub fn mono_to_stereo(x: Sample, panning: f32) -> Frame {
        let (l_amp, r_amp) = mono_pan_amplitudes(panning);

        Frame {l: l_amp * x, r: r_amp * x}
    }

    pub fn zero() -> Frame {
        Frame {l: 0., r: 0.,}
    }
}

impl Add for Frame {
    type Output = Self;
    fn add(self, rhs: Frame) -> Frame {
        Frame {
            l: self.l + rhs.l,
            r: self.r + rhs.r,
        }
    }
}

impl Mul<f32> for Frame {
    type Output = Self;
    fn mul(self, rhs: f32) -> Frame {
        Frame {
            l: self.l * rhs,
            r: self.r * rhs,
        }
    }
}

impl Sum for Frame {
    fn sum<I: Iterator<Item=Frame>>(iter: I) -> Frame {
        iter.fold(Frame::zero(), Add::add)
    }
}

pub trait Renderable {
    fn render_sample(&mut self, sample_rate: f64) -> Frame;
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn pan_amp() {
        let pan_right = mono_pan_amplitudes(0.2);
        assert!(pan_right.0 < pan_right.1);
        let pan_left = mono_pan_amplitudes(-0.2);
        assert!(pan_left.0 > pan_left.1);

        let pan_center = mono_pan_amplitudes(0.);
        assert!(pan_center.0 == pan_center.1);
    }

    #[test]
    fn bitpan() {
        assert_eq!(0., from_bitpan(0));
        assert_eq!((45.0_f32).to_radians(), from_bitpan(127));
        assert_eq!((-45.0_f32).to_radians(), from_bitpan(-127));
    }
}
